#include "PS2Mouse.h"
#include "elapsedMillis.h"
#include <Adafruit_SleepyDog.h>

/*
 * PS/2 to DEC 7-pin RS232 mouse converter
 * 
 * Adapted by Cathode Ray Dude (cathoderaydude@gmail.com)
 * Email with comments / suggestions / bugs.
 *
 * This is based on https://bitbucket.org/tinyscsi/decmouse/src/master/
 * I was unable to get that sketch to work on my Arduino "Pro Micro", whatever that is,
 * so this version has been significantly rewritten, and I switched to a different PS/2
 * mouse library that seems a lot more reliable.
 *
 * To make this work, you will need:
 * - An Arduino with at least one RS232 port. Not all of them have that! Did you know that! I used a Pro Micro.
 * - An RS232 level shifter to get the signal up to 12V - I'll assume you bought a preassembled MAX232 from e.g. adafruit with a DE9 plug.
 * - Adafruit SleepyDog library
 * 
 * Connect the level shifter to the TX/RX pins of the 'duino. TX goes to TX, RX to RX; no crossover.
 * Connect the level shifter's VCC and GND to the 'duino's matching pins.
 * Connect the CLOCK and DATA lines from the PS/2 mouse to any pair of free pins on the 'duino.
 * Update the CLOCK_PIN and DATA_PIN defines below with those pin numbers.
 * Connect the mouse's GND and VCC lines to same on the 'duino to give it power.
 *
 * Now all you have to do is build a DIN-7 (or -8, the sockets fit) to DE9 adapter.
 * Pin 1: GND   -> DE9 pin 5
 * Pin 2: TX    -> DE9 pin 2
 * Pin 3: RX    -> DE9 pin 3 (notice we DID cross over)
 * Pin 4: -12V  -> NC (be careful!!!)
 * Pin 5: +5V   -> NC (maybe you can power the 'duino from it...?)
 * Pin 6: NC
 * Pin 7: Device Present -> Short to GND
 *
 * Plug this into the MAX232, fire it all up, and if all goes well you should see the LED start blinking.
 *
 * 
 * You can actually hotplug PS/2 mice with this code. I mean, you can't, but you can.
 * If you unplug the mouse, the PS2Mouse library will go into a coma. Fortunately, Arduinos have watchdogs.
 * I've set it to 15 seconds, so if you unplug the mouse, after 15 seconds the 'duino will reset and
 * renegotiate.
 *
 * Note that if you do this, or otherwise reset the Arduino *while the machine is booted,* you aren't
 * going to have any mouse activity because it starts up in "D" state, e.g. "send me a command" mode.
 * This is necessary; otherwise it'll fail self-test and halt boot.
 *
 * Solution: If you have to hotplug the mouse while running, simply wait for the LED to start blinking
 * quickly (250ms) and then left-click. You should see the LED go out; the mouse is now in Stream Mode
 * and should start working.
 *
 *
 * Conversely, if you reset the *MACHINE* without power-cycling the Arduino, the mouse is going to fail
 * self-test; this can't be avoided. Mice are supposed to be powered off their host systems, so this
 * scenario should never occur in reality.
 * If your Arduino is powered off the DEC, you won't have this issue.
 * If you have it plugged into a separate source, you'll just have to reset it manually when rebooting.
 */

// Start in Poll Mode, e.g. waiting for a command
uint8_t DECmode = 'D';

/*
 * PS/2 mouse clock/data pins.
 * Set these to a pair of pins that are available on your particular Arduino.
 */
#define CLOCK_PIN 10
#define DATA_PIN 16
PS2Mouse mouse(CLOCK_PIN, DATA_PIN);

// status indicator LED
#define LED_PIN 17

void setup()
{
  // Turn on watchdog
  // From here onwards, any delays in e.g. the ps/2 routine will trip a reset
  // This allows you to hotplug mice! Waow
  int countdownMS = Watchdog.enable(15000);

  pinMode(LED_PIN, OUTPUT); // status LED = output
  digitalWrite(LED_PIN, HIGH);  // turn LED on
  
  // === IMPORTANT NOTE ===
  // this isn't 8,n,1! this is 8,Odd,1! yeah, I know you've never seen it before! me either! but it's crucial!!!
  // If you're testing by sending bytes from a PC serial port, half of them will arrive trashed
  // unless you set Odd Parity in putty et al. Ask me how I know.
  Serial1.begin(4800, SERIAL_8O1);
  
  // Wait for both RS232 and (potentially) Arduino IDE serial to come up.
  delay(1500);

  // Set up ps/2 mouse
  if (Serial) Serial.println("Detecting mouse");
  mouse.initialize(); // actually try to query the ps/2 mouse
  if (Serial) Serial.println("Mouse detected");

  // init complete, turn off light
  digitalWrite(LED_PIN, LOW);
}

elapsedMillis eLED;
int tLED = 0;

void loop()
{
  char mstat;
  char mx;
  char my;

  // Get PS/2 mouse status
  MouseData data = mouse.readData();
  mstat = data.status;
  my = data.position.y;
  mx = data.position.x;

  // Uncomment this if you need to debug whether the ps/2 mouse data
  // is being received in arduino IDE.
  /*
  MouseData data = mouse.readData();
  Serial.print(data.status, BIN);
  Serial.print("\tx=");
  Serial.print(data.position.x);
  Serial.print("\ty=");
  Serial.print(data.position.y);
  Serial.print("\twheel=");
  Serial.print(data.wheel);
  Serial.println();
  */

  // If controller is still in startup state and user left-clicks, jump into stream mode
  // This is for resetting while the host is still running.
  if ((char) DECmode == 'D' && (mstat & 0x1)) {
    if (Serial) Serial.println("Click - going to stream mode");
    DECmode = 'R';
  }

  // Status LED blinks every 250ms if still in startup state
  // During normal operation, off.
  if(eLED > 250 && DECmode == 'D') {
    eLED = 0;
    tLED = !tLED;
    digitalWrite(LED_PIN, tLED); // blinky
  } else {
    // Turn status LED off
    digitalWrite(LED_PIN, 1);
  }

  // Get mode command from DEC
  // Note that this is not sent on a regular basis, only when switching modes.
  // In practice: Only sent during boot.
  if(Serial1.available()) {
    uint8_t incomingByte = Serial1.read();
    switch(incomingByte) {
      case 'R':
        // 'R' is "continuously send position/button updates" mode.
        DECmode = 'R';
        if(Serial) Serial.print("Stream Mode\r\n");
        break;
      case 'P':
        // "send me a single position update"
        DECmode = 'P';
        if(Serial) Serial.print("Poll Read\r\n");
        break;
      case 'D':
        // "wait for instructions" - startup state for the mouse.
        DECmode = 'D';
        if(Serial) Serial.print("Poll Mode\r\n");
        break;
      case 'T':
        // "send a self-test response packet"
        DECmode = 'T';
        if(Serial) Serial.print("Self Test\r\n");
        break;
      default:
        // This means you're getting corrupt data.
        if(Serial) {
          Serial.print(incomingByte, HEX);
          Serial.print(" Unknown Byte\r\n");
        }
        break;
    }
  }

  // Self test requested
  if ((char) DECmode == 'T') {
      Serial1.write(0xA0);  // Self Test Report, REV0
      Serial1.write(0x02);  // Manufacturing ID 0, Mouse
      Serial1.write(0x00);  // No Errors
      Serial1.write(0x00);  // No Buttons Held
      
      // Once test is done, go back to D mode ("wait for command")
      if(DECmode == 'T')
        DECmode = 'D';
  }

  // Send position updates to DEC
  if ((char) DECmode == 'R' || (char) DECmode == 'P') {
      // magical bit shifting! don't change this!
      if(mx & 0x80) mx ^= 0x7f;
      if(my & 0x80) my ^= 0x7f;
      mx ^= 0x80;
      my ^= 0x80;

      // Write DEC Serial Mouse Packet
      Serial1.write(0x80 | ((mx & 0x80)>>3) | ((my & 0x80) >> 4) | ((mstat & 0x1) << 2) | (mstat & 0x2) | ((mstat & 0x4) >> 2));
      Serial1.write(mx & 0x7f);
      Serial1.write(my & 0x7f);

      // I believe "P" means "send a single packet, then stop," so if our last command
      // was a P, go back to D mode afterwards - in other words, wait for another explicit request.
      // However, if we were in R mode, we should just send another packet immediately.
      if(DECmode == 'P') DECmode = 'D';
  }

  // Feed the dog
  Watchdog.reset();

  // this is set to 5 here. huh. the ps2mouse library says a minimum of 20. i wonder if this should be changed.
  // to date I haven't had any problems.
  delay(5);
}


